Package: crisprBowtie
Version: 1.15.0
Date: 2025-07-22
Title: Bowtie-based alignment of CRISPR gRNA spacer sequences
Authors@R: c(
    person("Jean-Philippe", "Fortin", email = "fortin946@gmail.com", role = c("aut", "cre")))
Depends: methods
Imports: BiocGenerics, Biostrings, BSgenome, crisprBase (>= 0.99.15),
        Seqinfo, GenomicRanges, IRanges, Rbowtie, readr, stats,
        stringr, utils
Suggests: BiocStyle, BSgenome.Hsapiens.UCSC.hg38, knitr, rmarkdown,
        testthat
biocViews: CRISPR, FunctionalGenomics, Alignment
Description: Provides a user-friendly interface to map on-targets and off-targets
    of CRISPR gRNA spacer sequences using bowtie. The alignment is fast, 
    and can be performed using either commonly-used or custom CRISPR nucleases.
    The alignment can work with any reference or custom genomes.
    Both DNA- and RNA-targeting nucleases are supported.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.2.1
VignetteBuilder: knitr
BugReports: https://github.com/crisprVerse/crisprBowtie/issues
URL: https://github.com/crisprVerse/crisprBowtie
git_url: https://git.bioconductor.org/packages/crisprBowtie
git_branch: devel
git_last_commit: 5838b33
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-11-04
NeedsCompilation: no
Packaged: 2025-11-04 21:58:49 UTC; biocbuild
Author: Jean-Philippe Fortin [aut, cre]
Maintainer: Jean-Philippe Fortin <fortin946@gmail.com>

# Generated by roxygen2: do not edit by hand

export(runBowtie)
export(runCrisprBowtie)
importFrom(BSgenome,getSeq)
importFrom(BiocGenerics,end)
importFrom(BiocGenerics,start)
importFrom(Biostrings,DNAString)
importFrom(Biostrings,DNAStringSet)
importFrom(Biostrings,replaceLetterAt)
importFrom(Biostrings,reverseComplement)
importFrom(GenomicRanges,GRanges)
importFrom(IRanges,IRanges)
importFrom(IRanges,promoters)
importFrom(Rbowtie,bowtie)
importFrom(Seqinfo,seqlengths)
importFrom(Seqinfo,seqlevels)
importFrom(Seqinfo,seqnames)
importFrom(crisprBase,"spacerLength<-")
importFrom(crisprBase,extractPamFromTarget)
importFrom(crisprBase,extractProtospacerFromTarget)
importFrom(crisprBase,getTargetRanges)
importFrom(crisprBase,hasSpacerGap)
importFrom(crisprBase,isRnase)
importFrom(crisprBase,nucleaseName)
importFrom(crisprBase,pamIndices)
importFrom(crisprBase,pamLength)
importFrom(crisprBase,pamSide)
importFrom(crisprBase,pams)
importFrom(crisprBase,spacerLength)
importFrom(methods,is)
importFrom(readr,read_delim)
importFrom(stats,complete.cases)
importFrom(stringr,str_count)
importFrom(stringr,str_extract)
importFrom(utils,data)
importFrom(utils,write.table)

crisprBowtie: alignment of gRNA spacer sequences using bowtie
================

-   <a href="#overview-of-crisprbowtie"
    id="toc-overview-of-crisprbowtie">Overview of crisprBowtie</a>
-   <a href="#installation-and-getting-started"
    id="toc-installation-and-getting-started">Installation and getting
    started</a>
    -   <a href="#software-requirements" id="toc-software-requirements">Software
        requirements</a>
        -   <a href="#os-requirements" id="toc-os-requirements">OS Requirements</a>
    -   <a href="#installation-from-bioconductor"
        id="toc-installation-from-bioconductor">Installation from
        Bioconductor</a>
-   <a href="#building-a-bowtie-index"
    id="toc-building-a-bowtie-index">Building a bowtie index</a>
-   <a href="#alignment-using-runcrisprbowtie"
    id="toc-alignment-using-runcrisprbowtie">Alignment using
    <code>runCrisprBowtie</code></a>
-   <a href="#applications-beyond-crispr"
    id="toc-applications-beyond-crispr">Applications beyond CRISPR</a>
    -   <a href="#example-using-rnai-sirna-design"
        id="toc-example-using-rnai-sirna-design">Example using RNAi (siRNA
        design)</a>
-   <a href="#reproducibility" id="toc-reproducibility">Reproducibility</a>
-   <a href="#references" id="toc-references">References</a>

Authors: Jean-Philippe Fortin

Date: July 13, 2022

# Overview of crisprBowtie

`crisprBowtie` provides two main functions to align short DNA sequences
to a reference genome using the short read aligner bowtie (Langmead et
al. 2009) and return the alignments as R objects: `runBowtie` and
`runCrisprBowtie`. It utilizes the Bioconductor package `Rbowtie` to
access the Bowtie program in a platform-independent manner. This means
that users do not need to install Bowtie prior to using `crisprBowtie`.

The latter function (`runCrisprBowtie`) is specifically designed to map
and annotate CRISPR guide RNA (gRNA) spacer sequences using CRISPR
nuclease objects and CRISPR genomic arithmetics defined in the
Bioconductor package
[crisprBase](https://github.com/crisprVerse/crisprBase). This enables a
fast and accurate on-target and off-target search of gRNA spacer
sequences for virtually any type of CRISPR nucleases. It also provides
an off-target search engine for our main gRNA design package
[crisprDesign](https://github.com/crisprVerse/crisprDesign) of the
[crisprVerse](https://github.com/crisprVerse) ecosystem. See the
`addSpacerAlignments` function in `crisprDesign` for more details.

# Installation and getting started

## Software requirements

### OS Requirements

This package is supported for macOS, Linux and Windows machines. Package
was developed and tested on R version 4.2.1.

## Installation from Bioconductor

`crisprBowtie` can be installed from from the Bioconductor devel branch
using the following commands in a fresh R session:

``` r
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install(version="devel")
BiocManager::install("crisprBowtie")
```

The complete documentation for the package can be found
[here](https://bioconductor.org/packages/devel/bioc/manuals/crisprBowtie/man/crisprBowtie.pdf).

# Building a bowtie index

To use `runBowtie` or `runCrisprBowtie`, users need to first build a
Bowtie genome index. For a given genome, this step has to be done only
once. The `Rbowtie` package conveniently provides the function
`bowtie_build` to build a Bowtie index from any custom genome from a
FASTA file.

As an example, we build a Bowtie index for a small portion of the human
chromosome 1 (`chr1.fa` file provided in the `crisprBowtie` package) and
save the index file as `myIndex` to a temporary directory:

``` r
library(Rbowtie)
fasta <- file.path(find.package("crisprBowtie"), "example/chr1.fa")
tempDir <- tempdir()
Rbowtie::bowtie_build(fasta,
                      outdir=tempDir,
                      force=TRUE,
                      prefix="myIndex")
```

To learn how to create a Bowtie index for a complete genome or
transcriptome, please visit our [tutorial
page](https://github.com/crisprVerse/Tutorials/tree/master/Building_Genome_Indices).

# Alignment using `runCrisprBowtie`

As an example, we align 6 spacer sequences (of length 20bp) to the
custom genome built above, allowing a maximum of 3 mismatches between
the spacer and protospacer sequences.

We specify that the search is for the wildtype Cas9 (SpCas9) nuclease by
providing the `CrisprNuclease` object `SpCas9` available through the
`crisprBase` package. The argument `canonical=FALSE` specifies that
non-canonical PAM sequences are also considered (NAG and NGA for
SpCas9). The function `getAvailableCrisprNucleases` in `crisprBase`
returns a character vector of available `crisprNuclease` objects found
in `crisprBase`.

``` r
library(crisprBowtie)
data(SpCas9, package="crisprBase")
crisprNuclease <- SpCas9
spacers <- c("TCCGCGGGCGACAATGGCAT",
             "TGATCCCGCGCTCCCCGATG",
             "CCGGGAGCCGGGGCTGGACG",
             "CCACCCTCAGGTGTGCGGCC",
             "CGGAGGGCTGCAGAAAGCCT",
             "GGTGATGGCGCGGGCCGGGC")
runCrisprBowtie(spacers,
                crisprNuclease=crisprNuclease,
                n_mismatches=3,
                canonical=FALSE,
                bowtie_index=file.path(tempDir, "myIndex"))
```

    ## [runCrisprBowtie] Searching for SpCas9 protospacers

    ##                 spacer          protospacer pam  chr pam_site strand
    ## 1 CCACCCTCAGGTGTGCGGCC CCACCCTCAGGTGTGCGGCC TGG chr1      679      +
    ## 2 CCGGGAGCCGGGGCTGGACG CCGGGAGCCGGGGCTGGACG GAG chr1      466      +
    ## 3 CGGAGGGCTGCAGAAAGCCT CGGAGGGCTGCAGAAAGCCT TGG chr1      706      +
    ## 4 GGTGATGGCGCGGGCCGGGC GGTGATGGCGCGGGCCGGGC CGG chr1      831      +
    ## 5 TGATCCCGCGCTCCCCGATG TGATCCCGCGCTCCCCGATG CAG chr1      341      +
    ##   n_mismatches canonical
    ## 1            0      TRUE
    ## 2            0     FALSE
    ## 3            0      TRUE
    ## 4            0      TRUE
    ## 5            0     FALSE

# Applications beyond CRISPR

The function `runBowtie` is similar to `runCrisprBowtie`, but does not
impose constraints on PAM sequences. It can be used to search for any
short read sequence in a genome.

## Example using RNAi (siRNA design)

Seed-related off-targets caused by mismatch tolerance outside of the
seed region is a well-studied and characterized problem observed in RNA
interference (RNA) experiments. `runBowtie` can be used to map
shRNA/siRNA seed sequences to reference genomes to predict putative
off-targets:

``` r
seeds <- c("GTAAAGGT", "AAGGATTG")
runBowtie(seeds,
          n_mismatches=2,
          bowtie_index=file.path(tempDir, "myIndex"))
```

    ##      query   target  chr pos strand n_mismatches
    ## 1 AAGGATTG AAAGAATG chr1 163      -            2
    ## 2 AAGGATTG AAGCCTTG chr1 700      +            2
    ## 3 AAGGATTG AAGGCTTT chr1 699      -            2
    ## 4 AAGGATTG CAGGCTTG chr1 905      -            2
    ## 5 GTAAAGGT GGGAAGGT chr1 724      +            2

# Reproducibility

``` r
sessionInfo()
```

    ## R version 4.2.1 (2022-06-23)
    ## Platform: x86_64-apple-darwin17.0 (64-bit)
    ## Running under: macOS Catalina 10.15.7
    ## 
    ## Matrix products: default
    ## BLAS:   /Library/Frameworks/R.framework/Versions/4.2/Resources/lib/libRblas.0.dylib
    ## LAPACK: /Library/Frameworks/R.framework/Versions/4.2/Resources/lib/libRlapack.dylib
    ## 
    ## locale:
    ## [1] en_US.UTF-8/en_US.UTF-8/en_US.UTF-8/C/en_US.UTF-8/en_US.UTF-8
    ## 
    ## attached base packages:
    ## [1] stats     graphics  grDevices utils     datasets  methods   base     
    ## 
    ## other attached packages:
    ## [1] crisprBowtie_1.1.1 Rbowtie_1.37.0    
    ## 
    ## loaded via a namespace (and not attached):
    ##  [1] SummarizedExperiment_1.27.2 tidyselect_1.1.2           
    ##  [3] xfun_0.32                   purrr_0.3.4                
    ##  [5] lattice_0.20-45             vctrs_0.4.1                
    ##  [7] htmltools_0.5.3             stats4_4.2.1               
    ##  [9] rtracklayer_1.57.0          yaml_2.3.5                 
    ## [11] utf8_1.2.2                  XML_3.99-0.10              
    ## [13] rlang_1.0.5                 pillar_1.8.1               
    ## [15] glue_1.6.2                  BiocParallel_1.31.12       
    ## [17] bit64_4.0.5                 BiocGenerics_0.43.4        
    ## [19] matrixStats_0.62.0          GenomeInfoDbData_1.2.8     
    ## [21] lifecycle_1.0.1             stringr_1.4.1              
    ## [23] zlibbioc_1.43.0             MatrixGenerics_1.9.1       
    ## [25] Biostrings_2.65.3           codetools_0.2-18           
    ## [27] evaluate_0.16               restfulr_0.0.15            
    ## [29] Biobase_2.57.1              knitr_1.40                 
    ## [31] tzdb_0.3.0                  IRanges_2.31.2             
    ## [33] fastmap_1.1.0               GenomeInfoDb_1.33.7        
    ## [35] parallel_4.2.1              fansi_1.0.3                
    ## [37] crisprBase_1.1.8            readr_2.1.2                
    ## [39] BSgenome_1.65.2             DelayedArray_0.23.1        
    ## [41] S4Vectors_0.35.3            vroom_1.5.7                
    ## [43] XVector_0.37.1              bit_4.0.4                  
    ## [45] Rsamtools_2.13.4            rjson_0.2.21               
    ## [47] hms_1.1.2                   digest_0.6.29              
    ## [49] stringi_1.7.8               BiocIO_1.7.1               
    ## [51] GenomicRanges_1.49.1        grid_4.2.1                 
    ## [53] cli_3.4.0                   tools_4.2.1                
    ## [55] bitops_1.0-7                magrittr_2.0.3             
    ## [57] RCurl_1.98-1.8              tibble_3.1.8               
    ## [59] crayon_1.5.1                pkgconfig_2.0.3            
    ## [61] ellipsis_0.3.2              Matrix_1.4-1               
    ## [63] rmarkdown_2.16              rstudioapi_0.14            
    ## [65] R6_2.5.1                    GenomicAlignments_1.33.1   
    ## [67] compiler_4.2.1

# References

<div id="refs" class="references csl-bib-body hanging-indent">

<div id="ref-langmead2009bowtie" class="csl-entry">

Langmead, Ben, Cole Trapnell, Mihai Pop, and Steven L. Salzberg. 2009.
“Ultrafast and Memory-Efficient Alignment of Short DNA Sequences to the
Human Genome.” *Genome Biology* 10 (3): R25.
<https://doi.org/10.1186/gb-2009-10-3-r25>.

</div>

</div>

library(BSgenome.Hsapiens.UCSC.hg38)
library(Biostrings)
bsgenome=BSgenome.Hsapiens.UCSC.hg38
seq <- DNAStringSet(getSeq(bsgenome, "chr1", 1000000,1001000-1))
names(seq) <- "chr1"
writeXStringSet(seq, "chr1.fa")

#' @importFrom stats complete.cases
NULL

N_MAX_MISMATCHES <- 3

utils::globalVariables(c('.', "SpCas9", "AsCas12a"))


#' @importFrom methods is
.validateCrisprNuclease <- function(crisprNuclease){
    if (is.null(crisprNuclease)){
        crisprNuclease <- .getDefaultCrisprNuclease()
    } else {
        if (!is(crisprNuclease, "CrisprNuclease")){
            stop("Provided nuclease must be a 'CrisprNuclease' object. ")
        }
    }
    return(crisprNuclease)
}

#' @importFrom utils data
.getDefaultCrisprNuclease <- function(type=c("Cas9", "Cas12a")){
    type <- match.arg(type)
    if (type=="Cas9"){
        data("SpCas9",
             package="crisprBase",
             envir=environment())
        nuc <- SpCas9
    } else {
        data("AsCas12a",
             package="crisprBase",
             envir=environment())
        nuc <- AsCas12a
    }
    return(nuc)
}


#' @importFrom methods is
.validateBSGenome <- function(bsgenome){
    if (is.null(bsgenome)){
        stop("Provided bsgenome argument cannot be NULL.")
    } else {
        if (!is(bsgenome, "BSgenome")){
            stop("Provided bsgenome argument must be a 'BSgenome' object. ")
        }
    }
    return(bsgenome)
}




# Takes a character vector of sequences
# and writes to disk the sequences 
# in a fastq format. If temporary=TRUE,
# the fastq file is written in a 
# temporary folder. 
#' @importFrom utils write.table
.fastqfy <- function(sequences,
                     temporary=TRUE,
                     file=NULL
){
    lines <- list()
    lines[[1]] <- paste0("@", sequences)
    lines[[2]] <- sequences
    lines[[3]] <- paste0("+", sequences)
    lines[[4]] <- vapply(nchar(sequences), function(x){
                      paste0(rep('~', x), collapse='')
                  }, FUN.VALUE="a")    
    temp <- split(do.call(cbind, lines),
                  f=sequences)
    temp <- matrix(unlist(temp), ncol=1)
    if (temporary){
        file <- tempfile()
    } else {
        if (is.null(file)){
            stop("If temporary=FALSE, 'file' must be provided.")
        }
    }
    write.table(temp, 
                file=file,
                quote=FALSE,
                row.names=FALSE,
                col.names=FALSE)
    return(file)
}




# Takes a character vector of sequences
# and writes to disk the sequences 
# in a fasta format. If temporary=TRUE,
# the fasta file is written in a 
# temporary folder. 
#' @importFrom utils write.table
.fastafy <- function(sequences,
                     temporary=TRUE,
                     file=NULL
){
    lines <- list()
    lines[[1]] <- paste0(">", sequences)
    lines[[2]] <- sequences
    temp <- split(do.call(cbind, lines),
                  f=sequences)
    temp <- matrix(unlist(temp), ncol=1)
    if (temporary){
        file <- tempfile()
    } else {
        if (is.null(file)){
            stop("If temporary=FALSE, 'file' must be provided.")
        }
    }
    write.table(temp, 
                file=file,
                quote=FALSE,
                row.names=FALSE,
                col.names=FALSE)
    return(file)
}





.validateBowtieIndex <- function(bowtie_index){
    suffixes_fwd <- paste0(".", seq_len(4), ".ebwt")
    suffixes_rev <- paste0(".rev.", seq_len(2), ".ebwt")
    suffixes <- c(suffixes_fwd, suffixes_rev)
    files   <- paste0(bowtie_index, suffixes)
    missing <- files[!file.exists(files)]
    if (length(missing)>0){
        missing <- paste0(missing, collapse="\n") 
        stop("The following files needed for bowtie",
             " are missing: \n", missing)
    }
    return(bowtie_index)
}




#' @importFrom methods is
.checkBSGenomeOrNull <- function(bsgenome){
    if (!is.null(bsgenome)){
        if (!is(bsgenome, "BSgenome")){
            stop("Provided bsgenome argument must be a 'BSgenome' object or NULL. ")
        }
    }
    invisible(NULL)
}

#' @title Perform CRISPR gRNA spacer alignment with bowtie
#' @description Perform CRISPR gRNA spacer alignment with bowtie.
#' 
#' @param spacers Character vector specifying gRNA spacer sequences.
#'     Sequences must all be of equal length. 
#' @param mode String specifying which alignment mode should be used:
#'     \code{protospacer} or \code{spacer}. 
#'     For RNA-targeting nucleases such as CasRx, only the 
#'     protospacer mode can be used. 
#' @param bowtie_index String specifying path to a bowtie index.
#' @param bsgenome A \linkS4class{BSgenome} object.
#'     Must be provided if \code{mode} is "spacer".
#'     Ignore
#' @param crisprNuclease A \linkS4class{CrisprNuclease} object. 
#' @param canonical Should only canonical PAM sequences be considered?
#'     TRUE by default.
#' @param ignore_pam Should PAM sequences be ignore?
#'     If TRUE, all alignments are returned regardless of PAM tolerance.
#'     FALSE by default.
#' @param n_mismatches Integer between 0 and 3 specifying maximum number
#'     of mismatches allowed between spacer sequences and target DNA.
#'     0 by default. 
#' @param all_alignments Should all possible alignments be returned?
#'     TRUE by default. 
#' @param n_max_alignments Maximum number of alignments to return if
#'     \code{all_alignments} is FALSE. 1000 by default. 
#' @param force_spacer_length Should the spacer length be overwritten in the
#'     \code{crisprNuclease} object? FALSE by default. 
#' @param rna_strict_directionality Should only protospacers found in the 
#'     original direction of the RNA be considered for RNA-targeting
#'     nucleases? TRUE by default.
#' @param verbose Should messages be printed to the console? TRUE by default.
#' 
#' @return A data.frame of the spacer alignments with the following columns:
#'    \itemize{
#'        \item \code{spacer} — string specifying gRNA spacer sequence
#'        \item \code{protospacer} — string specifying target protospacer sequence
#'        \item \code{pam} — string specifying target PAM sequence
#'        \item \code{chr} - string specifying chromosome name
#'        \item \code{pam_site} - string specifying genomic coordinate of the
#'              first nucleotide of the PAM sequence.
#'        \item \code{strand} - string specifying strand ("+" or "-") 
#'        \item \code{n_mismatches} - integer specifying number of mismatches
#'              between spacer and protospacer sequences
#'        \item \code{canonical} - logical indicating whether or not PAM sequence
#'              is canonical. 
#'    }
#'     
#' 
#' @details When \code{mode} is "spacer", spacer sequences are aligned to the
#'     reference index without appending PAM sequences first. This requires the
#'     specification of a \linkS4class{BSgenome} object through the argument
#'     \code{bsgenome} to validate that the aligned spacer sequences are
#'     adjacent to valid PAM sequences. 
#'     
#'     When \code{mode} is "protospacer", sequences are aligned with all
#'     valid PAM sequences appended (spacer + PAM). The set of valid PAM
#'     sequences depend on the inputs  \code{canonical} and \code{ignore_pam}.
#'     This is faster than the "spacer" mode if the number of possible
#'     PAM sequences is small (e.g. SpCas9).
#' 
#'     For RNA-targeting nucleases, such as RfxCas13d (CasRx), the bowtie
#'     index should be built on a transcriptome. For such applications,
#'     only the "protospacer" mode can be used as there is no
#'     corresponding bsgenome package. The protospacer sequences
#'     searched in the reference index will be the reverse complement
#'     of the input spacer sequences.
#' 
#' @examples
#' fasta <- system.file(package="crisprBowtie", "example/chr1.fa")
#' outdir <- tempdir()
#' Rbowtie::bowtie_build(fasta,outdir=outdir, force=TRUE, prefix="tempIndex")
#' index <- file.path(outdir, "tempIndex")
#' seqs <- c("GGAAATTCCCCCAGTGGCGC",
#'           "ACACAGCTGCGGACAGGGCC")
#' data(SpCas9, package="crisprBase")
#' results <- runCrisprBowtie(seqs,
#'                            bowtie_index=index,
#'                            n_mismatches=2,
#'                            crisprNuclease=SpCas9)
#' 
#' @seealso \code{\link{runBowtie}} to map general DNA sequences.
#' 
#' @author Jean-Philippe Fortin
#' 
#' @export
#' @importFrom Seqinfo seqlengths seqnames
#' @importFrom BSgenome getSeq
#' @importFrom crisprBase extractPamFromTarget
#' @importFrom crisprBase extractProtospacerFromTarget
#' @importFrom crisprBase nucleaseName pams pamLength pamIndices
#' @importFrom crisprBase spacerLength spacerLength<- pamSide
#' @importFrom crisprBase hasSpacerGap isRnase
#' @importFrom crisprBase getTargetRanges
#' @importFrom Biostrings reverseComplement DNAStringSet
#' @importFrom BiocGenerics start end
runCrisprBowtie <- function(spacers, 
                            mode=c("protospacer", "spacer"),
                            bowtie_index=NULL,
                            bsgenome=NULL,
                            crisprNuclease=NULL,
                            canonical=TRUE,
                            ignore_pam=FALSE,
                            n_mismatches=0, 
                            all_alignments=TRUE,
                            n_max_alignments=1000,
                            force_spacer_length=FALSE,
                            rna_strict_directionality=TRUE,
                            verbose=TRUE
){ 
    #Checking inputs:
    mode     <- match.arg(mode)
    crisprNuclease <- .validateCrisprNuclease(crisprNuclease)
    if (hasSpacerGap(crisprNuclease)){
        stop("CRISPR nucleases with spacer gaps are not ",
             "supported at the moment.")
    }
    if (isRnase(crisprNuclease) & mode=="spacer"){
        stop("For RNA-targeting nucleases, the alignment mode must ",
             "be set to 'protospacer'.")
    }
    if (mode=="spacer"){
        bsgenome <- .validateBSGenome(bsgenome)
    }
    

    if (is.null(bowtie_index)){
        stop("bowtie_index must be provided.")
    }
    bowtie_index <- .validateBowtieIndex(bowtie_index)
    

    # Checking mode arguments:
    if (mode=="spacer" & is.null(bsgenome)){
        stop("In 'spacer' mode, a BSgenome object must be provided using",
             " the argument 'bsgenome'.")
    }
    if (mode=="spacer" & verbose){
        cat(paste0("[runCrisprBowtie] Using ", bsgenome@pkgname, " \n"))
    }
    if (verbose){
        cat(paste0("[runCrisprBowtie] Searching for ",
                   nucleaseName(crisprNuclease), " protospacers \n"))
    }

    # Getting nuclease info:
    pam.len     <- pamLength(crisprNuclease)
    pam.side    <- pamSide(crisprNuclease) 
    spacer.len  <- unique(nchar(spacers))

    if (length(spacer.len)>1){
        stop("Spacers must all have the same length.")
    }
    spacer.len.from.nuclease <- spacerLength(crisprNuclease)
    if (!force_spacer_length){
        if (spacer.len.from.nuclease != spacer.len){
            stop("Length of provided spacers is ",
                 spacer.len,
                 ", but spacer length for the provided nuclease is ",
                 spacer.len.from.nuclease, 
                 ". Consider force_spacer_length=TRUE to overwrite the",
                 " nuclease spacer length. ")
        }
    } else {
        if (spacer.len.from.nuclease != spacer.len){
            message("Setting spacer length to be ", spacer.len)
            spacerLength(crisprNuclease) <- spacer.len
        }
    }
    
    possiblePams      <- pams(crisprNuclease, 
                              primary=canonical,
                              ignore_pam=ignore_pam,
                              as.character=TRUE)
    pams.canonical    <- pams(crisprNuclease,
                              primary=TRUE,
                              as.character=TRUE)
    pams.noncanonical <- pams(crisprNuclease,
                              primary=FALSE,
                              as.character=TRUE)
    pam.indices <- pamIndices(crisprNuclease)



    # Getting input sequences:
    if (mode=="spacer"){
        sequences <- spacers
    } else {
        if (isRnase(crisprNuclease)){
            sequences <- reverseComplement(DNAStringSet(spacers))
            sequences <- as.character(sequences)
        } else {
            sequences <- spacers
        }

        if (pam.side=="3prime"){
            protospacers <- lapply(possiblePams, function(x){
                paste0(sequences, x)
            })
        } else {
            protospacers <- lapply(possiblePams, function(x){
                paste0(x, sequences)
            })
        }
        sequences <- do.call("c", protospacers)
    }

    # Performing alignment:
    aln <- runBowtie(sequences=sequences, 
                     bowtie_index=bowtie_index, 
                     bsgenome=bsgenome,
                     n_mismatches=n_mismatches,
                     n_max_alignments=n_max_alignments,
                     all_alignments=all_alignments,
                     verbose=verbose)
    if (is.null(aln)){
        return(.emptyAlignments())
    }

    #cat("Getting pam site \n")
    aln$pam_site <- .getPamSiteFromBowtieOutput(pos=aln$pos, 
                                                strand=aln$strand, 
                                                spacer.len=spacer.len,
                                                crisprNuclease=crisprNuclease,
                                                mode=mode)
    aln <- aln[aln$pam_site>0,,drop=FALSE]
    if (nrow(aln) == 0){
        return(.emptyAlignments())
    }

    #cat("Getting PAM sequences \n")
    if (mode == "spacer"){
        if (nrow(aln) > 0){
            # Filtering out PAMs falling outside of chrs
            protoRanges <- getTargetRanges(seqnames=aln$chr,
                                           pam_site=aln$pam_site,
                                           strand=aln$strand,
                                           nuclease=crisprNuclease)
            chr_lens <- seqlengths(bsgenome)[as.character(seqnames(protoRanges))]
            valid <- BiocGenerics::start(protoRanges) > 0 &
                BiocGenerics::end(protoRanges) <= chr_lens
            if (any(is.na(valid))){
                wMessage <- paste("Some alignments found in seqlevels not",
                                  "present in bsgenome, ignoring.")
                warning(wMessage,
                        immediate.=TRUE)
                valid[is.na(valid)] <- FALSE
            }
            aln <- aln[valid,,drop=FALSE]
        }
        if (nrow(aln) > 0){
            pam.gr <- .getBowtiePamRanges(chr=aln$chr,
                                          pam_site=aln$pam_site,
                                          strand=aln$strand,
                                          crisprNuclease=crisprNuclease)
            aln$pam <- as.character(getSeq(bsgenome, pam.gr))
            aln$canonical <- aln$pam %in% pams.canonical
        } 
    } else {
        if (nrow(aln) > 0){
            aln$pam    <- extractPamFromTarget(aln$target,
                                                    crisprNuclease)
            aln$target <- extractProtospacerFromTarget(aln$target,
                                                       crisprNuclease)
            aln$query  <- extractProtospacerFromTarget(aln$query,
                                                       crisprNuclease)
            aln$canonical <- aln$pam %in% pams.canonical
        } 
        # To remove mismatches occuring in pam sequences:
        # Ordering by number of mismatches is important
        # as higher number of mismatches for identical 
        # spacer and protospacer indicates that the 
        # last-ordered row has a mismatch in the PAM
        aln <- aln[order(aln$query,
                         aln$target,
                         aln$chr,
                         aln$pam_site,
                         aln$n_mismatches),,drop=FALSE]

        ids <- paste0(aln$query, aln$target, aln$chr, aln$pam_site)
        bad <- which(duplicated(ids))
        #bad <- which(aln$mm1 %in% pam.indices|
        #             aln$mm2 %in% pam.indices| 
        #             aln$mm3 %in% pam.indices)
        if (length(bad) > 0){
            aln <- aln[-bad,,drop=FALSE]
        }
    }
    if (canonical & !ignore_pam){
        aln <- aln[aln$canonical,,drop=FALSE]
    } else if (!canonical & !ignore_pam){
        aln  <- aln[aln$pam %in% pams.noncanonical,,drop=FALSE]
    }


    if (nrow(aln)==0){
        aln <- .emptyAlignments()
    } else {
        aln <- aln[order(aln$query, aln$n_mismatches),,drop=FALSE]
        aln$pos <- NULL
        colnames(aln)[colnames(aln)=="query"]  <- "spacer"
        colnames(aln)[colnames(aln)=="target"] <- "protospacer"
        aln <- aln[ ,.outputColumns(), drop=FALSE]
        rownames(aln) <- NULL
    }

    # RNAse considerations:
    if (isRnase(crisprNuclease)){
        spacers <- reverseComplement(DNAStringSet(aln$spacer))
        aln$spacer <- as.character(spacers)
        if (rna_strict_directionality){
            aln <- aln[aln$strand=="+",,drop=FALSE]
        }
    }

    return(aln)
}



.outputColumns <- function(){
     cols <- c("spacer",
               "protospacer",
               "pam",
               "chr",
               "pam_site",
               "strand", 
               "n_mismatches", 
               "canonical")
    return(cols)
}


.emptyAlignments <- function(){
    cols <- .outputColumns()
    out <- data.frame(matrix(0, ncol=length(cols)))
    colnames(out) <- cols
    out <- out[-1,,drop=FALSE]
    return(out)
}


#' @importFrom crisprBase pamLength pamSide
.getPamSiteFromBowtieOutput <- function(pos, 
                                        strand,
                                        spacer.len,
                                        crisprNuclease=NULL,
                                        mode=c("spacer", "protospacer")
){
    mode <- match.arg(mode)
    crisprNuclease <- .validateCrisprNuclease(crisprNuclease)
    pam.len  <- pamLength(crisprNuclease)
    pam.side <- pamSide(crisprNuclease)
    wh <- which(strand=="-")
    if (pam.side=="3prime"){
        pam_site <- pos + spacer.len
        if (length(wh)>0){
            pam_site[wh] <- pos[wh] - 1
            if (mode=="protospacer"){
                pam_site[wh] <- pam_site[wh] + pam.len
            }
        }
    } else {
        pam_site <- pos - pam.len
        if (mode=="protospacer"){
            pam_site <- pos
        }
        if (length(wh)>0){
            pam_site[wh] <- pos[wh] +spacer.len + pam.len - 1
        }
    }
    return(pam_site)
}




#' @importFrom GenomicRanges GRanges
#' @importFrom IRanges IRanges promoters
#' @importFrom crisprBase pamLength
.getBowtiePamRanges <- function(chr,
                                pam_site,
                                strand,
                                crisprNuclease=NULL
){
    crisprNuclease <- .validateCrisprNuclease(crisprNuclease)
    pam.len  <- pamLength(crisprNuclease)
    # print(pam_site)
    # print(strand)
    gr <- GRanges(chr, IRanges(pam_site, width=1), strand=strand)
    gr <- promoters(gr, downstream=pam.len, upstream=0)
    return(gr)
}

context('Testing runCrisprBowtie functions')

library(crisprBase)
library(Rbowtie)

# Creating bowtie index:
fasta  <- system.file(package="crisprBowtie", "example/chr1.fa")
outdir <- tempdir()
Rbowtie::bowtie_build(fasta,
                      outdir=outdir,
                      force=TRUE,
                      prefix="tempIndex")
index <- file.path(outdir, "tempIndex")

data(SpCas9, package="crisprBase")
data(AsCas12a, package="crisprBase")
Cas9 <- SpCas9
Cas12a <- AsCas12a


spacers_cas9 <- c("GGAAATTCCCCCAGTGGCGC",
                  "GGAACTTCCCCCCGTGGCGC",
                  "GGAGCACGGAGGGCTGCAGA",
                  "GCGGGCGAGCGGCGAGCGCG",
                  "GCGGGCGAGCAGCGAGGGCG",
                  "CGCCCTCCCGCCTGCCGACC",
                  "GGGCCTCACGCCTGCCGACC")

spacers_cas12a <- c("AGGCCTCGCCGCCGCGGGGTGTG",
                    "CCACACTCGCCCCAGCCAATCGA",
                    "TGCAGCCCTCCGTGCTCCAGGCC",
                    "TGCAGCCCTCGGTGCTCCAGGCC",
                    "TGCAGCCCTCGGTACTCCAGGCC",
                    "TGCAGCCCTCGGTACTGCAGGCC")



test_that('Testing runCrisprBowtie with Cas9 canonical', {
    results_cas9_mm0 <- runCrisprBowtie(spacers_cas9,
                                        bowtie_index=index,
                                        n_mismatches=0,
                                        crisprNuclease=Cas9,
                                        canonical=TRUE)
    results_cas9_mm1 <- runCrisprBowtie(spacers_cas9,
                                        bowtie_index=index,
                                        n_mismatches=1,
                                        crisprNuclease=Cas9,
                                        canonical=TRUE)
    results_cas9_mm2 <- runCrisprBowtie(spacers_cas9,
                                        bowtie_index=index,
                                        n_mismatches=2,
                                        crisprNuclease=Cas9,
                                        canonical=TRUE)
    results_cas9_mm3 <- runCrisprBowtie(spacers_cas9,
                                        bowtie_index=index,
                                        n_mismatches=3,
                                        crisprNuclease=Cas9,
                                        canonical=TRUE)
    update <- FALSE 
    expect_equal_to_reference(results_cas9_mm0,
                              file=file.path("objects/results_cas9_mm0.rds"),
                              update=update)
    expect_equal_to_reference(results_cas9_mm1,
                              file=file.path("objects/results_cas9_mm1.rds"),
                              update=update)
    expect_equal_to_reference(results_cas9_mm2,
                              file=file.path("objects/results_cas9_mm2.rds"),
                              update=update)
    expect_equal_to_reference(results_cas9_mm3,
                              file=file.path("objects/results_cas9_mm3.rds"),
                              update=update)
})


test_that('Testing runCrisprBowtie with Cas9 non-canonical', {
    results_cas9_mm0_nc <- runCrisprBowtie(spacers_cas9,
                                        bowtie_index=index,
                                        n_mismatches=0,
                                        crisprNuclease=Cas9,
                                        canonical=FALSE)
    results_cas9_mm1_nc <- runCrisprBowtie(spacers_cas9,
                                        bowtie_index=index,
                                        n_mismatches=1,
                                        crisprNuclease=Cas9,
                                        canonical=FALSE)
    results_cas9_mm2_nc <- runCrisprBowtie(spacers_cas9,
                                        bowtie_index=index,
                                        n_mismatches=2,
                                        crisprNuclease=Cas9,
                                        canonical=FALSE)
    results_cas9_mm3_nc <- runCrisprBowtie(spacers_cas9,
                                        bowtie_index=index,
                                        n_mismatches=3,
                                        crisprNuclease=Cas9,
                                        canonical=FALSE)
    update <- FALSE 
    expect_equal_to_reference(results_cas9_mm0_nc,
                              file=file.path("objects/results_cas9_mm0_nc.rds"),
                              update=update)
    expect_equal_to_reference(results_cas9_mm1_nc,
                              file=file.path("objects/results_cas9_mm1_nc.rds"),
                              update=update)
    expect_equal_to_reference(results_cas9_mm2_nc,
                              file=file.path("objects/results_cas9_mm2_nc.rds"),
                              update=update)
    expect_equal_to_reference(results_cas9_mm3_nc,
                              file=file.path("objects/results_cas9_mm3_nc.rds"),
                              update=update)
})


spacers_cas9_short <- c("AATTCCCCCAGTGGCGC",
                        "ACTTCCCCCCGTGGCGC",
                        "GCACGGAGGGCTGCAGA",
                        "GGCGAGCGGCGAGCGCG",
                        "GGCGAGCAGCGAGGGCG",
                        "CCTCCCGCCTGCCGACC",
                        "CCTCACGCCTGCCGACC")


test_that('Testing runCrisprBowtie with Cas9 (short spacers)', {
    expect_error(runCrisprBowtie(spacers_cas9_short,
                                 bowtie_index=index,
                                 n_mismatches=0,
                                 crisprNuclease=Cas9,
                                 canonical=FALSE))
    update <- FALSE 
    results_short_cas9 <- runCrisprBowtie(spacers_cas9_short,
                                          bowtie_index=index,
                                          n_mismatches=3,
                                          crisprNuclease=Cas9,
                                          canonical=FALSE,
                                          force_spacer_length=TRUE)
    expect_equal_to_reference(results_short_cas9,
                              file=file.path("objects/results_short_cas9.rds"),
                              update=update)

})



test_that('Testing runCrisprBowtie with Cas12a', {
                      
    results_cas12a_mm0 <- runCrisprBowtie(spacers_cas12a,
                                          bowtie_index=index,
                                          n_mismatches=0,
                                          crisprNuclease=Cas12a,
                                          canonical=TRUE)
    results_cas12a_mm1 <- runCrisprBowtie(spacers_cas12a,
                                          bowtie_index=index,
                                          n_mismatches=1,
                                          crisprNuclease=Cas12a,
                                          canonical=TRUE)
    results_cas12a_mm2 <- runCrisprBowtie(spacers_cas12a,
                                          bowtie_index=index,
                                          n_mismatches=2,
                                          crisprNuclease=Cas12a,
                                          canonical=TRUE)
    results_cas12a_mm3 <- runCrisprBowtie(spacers_cas12a,
                                          bowtie_index=index,
                                          n_mismatches=3,
                                          crisprNuclease=Cas12a,
                                          canonical=TRUE)
    update <- FALSE 
    expect_equal_to_reference(results_cas12a_mm0,
                              file=file.path("objects/results_cas12a_mm0.rds"),
                              update=update)
    expect_equal_to_reference(results_cas12a_mm1,
                              file=file.path("objects/results_cas12a_mm1.rds"),
                              update=update)
    expect_equal_to_reference(results_cas12a_mm2,
                              file=file.path("objects/results_cas12a_mm2.rds"),
                              update=update)
    expect_equal_to_reference(results_cas12a_mm3,
                              file=file.path("objects/results_cas12a_mm3.rds"),
                              update=update)
})


spacers_cas12a_short <- c("AGGCCTCGCCGCCGCGGGGT",
                          "CCACACTCGCCCCAGCCAAT",
                          "TGCAGCCCTCCGTGCTCCAG",
                          "TGCAGCCCTCGGTGCTCCAG",
                          "TGCAGCCCTCGGTACTCCAG",
                          "TGCAGCCCTCGGTACTGCAG")


test_that('Testing runCrisprBowtie with Cas12a (short spacers)', {
    expect_error(runCrisprBowtie(spacers_cas12a_short,
                                 bowtie_index=index,
                                 n_mismatches=0,
                                 crisprNuclease=Cas12a,
                                 canonical=FALSE))
    update <- FALSE 
    results_short_cas12a <- runCrisprBowtie(spacers_cas12a_short,
                                            bowtie_index=index,
                                            n_mismatches=3,
                                            crisprNuclease=Cas12a,
                                            canonical=FALSE,
                                            force_spacer_length=TRUE)
    expect_equal_to_reference(results_short_cas12a,
                              file=file.path("objects/results_short_cas12a.rds"),
                              update=update)
})

## ----eval=FALSE---------------------------------------------------------------
# if (!require("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install(version="devel")
# BiocManager::install("crisprBowtie")

## -----------------------------------------------------------------------------
library(Rbowtie)
fasta <- file.path(find.package("crisprBowtie"), "example/chr1.fa")
tempDir <- tempdir()
Rbowtie::bowtie_build(fasta,
                      outdir=tempDir,
                      force=TRUE,
                      prefix="myIndex")

## ----warning=FALSE, message=FALSE---------------------------------------------
library(crisprBowtie)
data(SpCas9, package="crisprBase")
crisprNuclease <- SpCas9
spacers <- c("TCCGCGGGCGACAATGGCAT",
             "TGATCCCGCGCTCCCCGATG",
             "CCGGGAGCCGGGGCTGGACG",
             "CCACCCTCAGGTGTGCGGCC",
             "CGGAGGGCTGCAGAAAGCCT",
             "GGTGATGGCGCGGGCCGGGC")
runCrisprBowtie(spacers,
                crisprNuclease=crisprNuclease,
                n_mismatches=3,
                canonical=FALSE,
                bowtie_index=file.path(tempDir, "myIndex"))

## ----eval=TRUE----------------------------------------------------------------
seeds <- c("GTAAAGGT", "AAGGATTG")
runBowtie(seeds,
          n_mismatches=2,
          bowtie_index=file.path(tempDir, "myIndex"))

## -----------------------------------------------------------------------------
sessionInfo()

library(crisprBowtie)
library(crisprBase)
library(Biostrings)
data(CasRx, package="crisprBase")
crisprNuclease <- CasRx

spacers <- c("GCCTTGAGAGCACCTCCTACACG",
             "GGAGAATTCCCCTACTCCTGCAT",
             "CTCCTTCTTCCTTTCCAGGCTTT",
             "CTGTCGTGCAGTCTCTTCCACAT")

bowtie_index <- "/Users/fortinj2/crisprIndices/bowtie/ensembl_human_104/ensembl_human_104"
aln <- runCrisprBowtie(spacers,
                       crisprNuclease=crisprNuclease,
                       bowtie_index=bowtie_index,
                       n_mismatches=2,
                       canonical=FALSE)

aln

#' fasta <- system.file(package="crisprBowtie", "example/chr1.fa")
#' outdir <- tempdir()
#' Rbowtie::bowtie_build(fasta,outdir=outdir, force=TRUE, prefix="tempIndex")
# index <- "/Users/fortinj2/crisprIndices/bowtie/hg38/hg38"
# bowtie_index <- index
# sequences <- c("GGAAGTAAATTTGGT",
#           "GTGGACCTCGGGAAT",
#           "GTGTGCGGTAAAGTC")
# n_mismatches=2
# results <- runBowtie(seqs,
#                      bowtie_index=index,
#                      n_mismatches=2)


# library(BSgenome.Hsapiens.UCSC.hg38)
# bsgenome <- BSgenome.Hsapiens.UCSC.hg38



#' @title Perform short sequence alignment with bowtie 
#' @description Perform short sequence alignment with bowtie. 
#' 
#' @param sequences Character vector of DNA sequences.
#' @param bowtie_index String specifying path to a bowtie index.
#' @param bsgenome BSgenome object.
#' @param n_mismatches Integer between 0 and 3 specifying maximum number
#'     of mismatches allowed between query sequences and target DNA.
#'     0 by default. 
#' @param all_alignments Should all possible alignments be returned?
#'     TRUE by default. 
#' @param n_max_alignments Maximum number of alignments to return if
#'     \code{all_alignments} is FALSE. 1000 by default. 
#' @param verbose Should messages be printed to the console?
#'     TRUE by default.
#' 
#' @return A data.frame of the alignments with the following columns:
#'    \itemize{
#'        \item \code{query} — string specifying query DNA sequence
#'        \item \code{target} — string specifying target DNA sequence
#'        \item \code{chr} - string specifying chromosome name
#'        \item \code{pos} - string specifying genomic coordinate of the start
#'              of the target DNA sequence
#'        \item \code{strand} - string specifying strand ("+" or "-") 
#'        \item \code{n_mismatches} - integer specifying number of mismatches
#'              between query and target sequences
#'    }
#'     
#' 
#' @details \code{runBowtie} can be used to map short DNA sequences 
#'     to a reference genome. To search for sequences while imposing
#'     constraints on PAM sequences (such as gRNA spacer sequences), see
#'     \code{runCrisprBowtie} instead.  
#' 
#' @examples
#' fasta <- system.file(package="crisprBowtie", "example/chr1.fa")
#' outdir <- tempdir()
#' Rbowtie::bowtie_build(fasta,outdir=outdir, force=TRUE, prefix="tempIndex")
#' index <- file.path(outdir, "tempIndex")
#' seqs <- c("GGAAGT",
#'           "GTGGAC",
#'           "GTGTGC") 
#' results <- runBowtie(seqs, bowtie_index=index, n_mismatches=2)
#' 
#' 
#' @seealso \code{\link{runCrisprBowtie}} to map gRNA spacer sequences.
#' 
#' @author Jean-Philippe Fortin
#' 
#' @importFrom Biostrings DNAStringSet DNAString replaceLetterAt
#' @importFrom Rbowtie bowtie
#' @export
runBowtie <- function(sequences, 
                      bowtie_index,
                      bsgenome=NULL,
                      n_mismatches=0,
                      all_alignments=TRUE,
                      n_max_alignments=1000,
                      verbose=TRUE
){    
    .checkNMismatches(n_mismatches)
    .checkBSGenomeOrNull(bsgenome)
    # Note: bowtie is based on python,
    # so index is 0-based instead of 1-based
    sequences <- unique(sequences)
    if (is.null(bowtie_index)){
        stop("bowtie_index must be provided.")
    }
    bowtie_index <- .validateBowtieIndex(bowtie_index)
                                         
    
    # Generating alignments:
    input <- .fastafy(sequences,
                      temporary=TRUE)
    outfile <- tempfile()
    results <- bowtie(sequences=input,
                      type="single",
                      index=bowtie_index,
                      outfile=outfile,
                      f=TRUE,
                      v=n_mismatches,
                      a=all_alignments,
                      k=crisprBase:::.makeLongInteger(n_max_alignments), 
                      force=TRUE)
    #cat("Reading bowtie output file. \n")
    results <- .readBowtieResults(outfile)



    if (file.exists(input)){
        file.remove(input)
    }
    # In case of no alignments:
    if (length(results)==0){ 
        return(NULL) 
    }  

    #cat("Getting DNA target \n")
    if (is.null(bsgenome)){
        results <- .getDNATargetFromMismatches(results, sequences)
    } else {
        #cat("Getting target sequences \n")
        results <- .getDNATargetFromBSgenome(results, bsgenome)
    }

    # Cleaning up:
    cols <- c("query", "target",
              "chr", "pos", "strand",
              "n_mismatches")
    results <- results[, cols, drop=FALSE]
    results <- results[order(results$query,
                             results$target,
                             results$chr, 
                             results$pos, 
                             results$strand,
                             results$n_mismatches),,drop=FALSE]
    rownames(results) <- NULL
    return(results)
}




#' @importFrom readr read_delim
#' @importFrom stringr str_count
.readBowtieResults <- function(file){
    isEmpty <- file.size(file)==0L
    if (!isEmpty){
        cols <- c("query", "strand", "chr", "pos",
                  "target", "qc", "idk", "mismatches")
        results <- read_delim(file,
                              col_names=cols,
                              col_types=c("ccciccic"),
                              col_select=c(1,2,3,4,8),
                              delim="\t") 
        results <- as.data.frame(results)
        cols  <- c("query", "chr", "pos", "strand", "mismatches")    
        results <- results[,cols,drop=FALSE]
        results$mismatches[is.na(results$mismatches)] <- ""
        results$pos <- results$pos+1 #since bowtie is 0-based
        results$n_mismatches <- str_count(results$mismatches, "\\:")
    } else {
        results <- vector(length=0)
    }
    return(results)
}





.getDNATargetFromMismatches <- function(results, sequences){
    
    # Processing mismatch info:
    results$mismatches <- as.character(results$mismatches)
    results$mismatches[results$mismatches==""]  <- "none"
    results$mismatches[is.na(results$mismatches)] <- "none"
    mismatches <- .processBowtieMismatches(results$mismatches)
    results    <- data.frame(results,
                             mismatches,
                             stringsAsFactors=FALSE)
    results$mismatches   <- NULL
    mm_cols <- paste0("mm", seq_len(N_MAX_MISMATCHES))
    

    # Complementing mismatches for negative strand:
    wh.neg <- which(results$strand=="-")
    if (length(wh.neg)>0){
        for (i in seq_len(N_MAX_MISMATCHES)){
            col <- paste0("mmnuc", i)
            .complement <- function(x){
                chartr("ATGC","TACG",x)
            }
            results[[col]][wh.neg] <- .complement(results[[col]][wh.neg])  
        }
    }

    # Getting target sequences in the absence of names:
    if (is.numeric(results$query)){
        results$query  <- sequences[as.integer(results$query)+1]
    }
    results$target <- results$query

    # Getting target genomic DNA sequence:
    sameLength <- length(unique(nchar(results$target)))==1
    if (sameLength){
        # Fast replacement with rectangular array:
        gRNA <- as.matrix(DNAStringSet(results$target))
        for (k in seq_len(N_MAX_MISMATCHES)){
            i <- seq_len(nrow(gRNA))
            j <- as.numeric(results[, paste0("mm",k)])
            x <- as.character(results[, paste0("mmnuc",k)])
            ind <- cbind(i,j)
            wh.complete <- complete.cases(ind)
            ind <- ind[wh.complete,,drop=FALSE]
            x <- x[wh.complete]
            if (length(x)>0){
                gRNA[ind] <- x
            }
        }
        DNA <- apply(gRNA, 1, paste, collapse="")
    } else {
        # Slow replacement with query sequences of different lengths:
        gRNAs <- lapply(results$target, DNAString)
        for (k in seq_len(N_MAX_MISMATCHES)){
            pos    <- as.numeric(results[, paste0("mm",k)])
            letter <- as.numeric(results[, paste0("mmnuc",k)])
            good <- which(!is.na(letter))
            if (length(good)>0){
                for (wh in good){
                    gRNAs[[wh]] <- replaceLetterAt(gRNAs[[wh]],
                                                   at=pos[wh],
                                                   letter=letter[wh])
                }
            }
        }
        DNA  <- as.character(DNAStringSet(gRNAs))
    }
    results$target <- DNA

    return(results)
}



#' @importFrom BSgenome getSeq
#' @importFrom Seqinfo seqlevels
.getDNATargetFromBSgenome <- function(results, bsgenome){

    validSeqLevels <- seqlevels(bsgenome)
    if (sum(results$chr %in% validSeqLevels)==0){
        stop("The chr info in the bowtie index does not match ",
             "the chr info in the bsgenome object.")
    }
    if (any(!results$chr %in% validSeqLevels)){
        wMessage <- paste("Some alignments map to chromosome names not found",
                          "in the BSgenome object; those alignments will be",
                          "ignored.")
        warning(wMessage,
                immediate.=TRUE)
    }
    results <- results[results$chr %in% validSeqLevels,,drop=FALSE]

    dna <- BSgenome::getSeq(bsgenome,
                            names=results$chr,
                            start=results$pos,
                            end=results$pos+nchar(results$query)-1,
                            strand=results$strand,
                            as.character=TRUE)
    results$target <- dna
    return(results)
}




.checkNMismatches <- function(n_mismatches){
    if (n_mismatches>3){
        stop("Number of mismatches must be either 0, 1, 2 or 3.")
    }
}

.processBowtieMismatches <- function(mismatches){
    out <- data.frame(.processMismatchPositions(mismatches),
                      .processMismatchLetters(mismatches),
                      stringsAsFactors=FALSE,
                      check.names=FALSE)
    return(out)
}


#' @importFrom stringr str_extract
.processMismatchPositions <- function(mismatches){
    mismatches <- strsplit(mismatches, split=",") 
    
    # Getting mismatch positions:
    mismatches.pos <- lapply(mismatches, function(x){
        x <- as.numeric(str_extract(x, "[0-9]+"))
        x <- x + 1 #since bowtie are 0-based
        for (i in seq_len(N_MAX_MISMATCHES-1)){
            if (length(x)==i){
                x <- c(x, rep(NA, N_MAX_MISMATCHES-i))
                break
            }
        }
        return(x)
    })
    mismatches.pos     <- do.call("rbind", mismatches.pos)
    for (i in seq_len(N_MAX_MISMATCHES)){
        mismatches.pos[,i] <- as.numeric(mismatches.pos[,i])  
    }
    colnames(mismatches.pos) <- paste0("mm",
                                       seq_len(N_MAX_MISMATCHES))
    return(mismatches.pos)
}


#' @importFrom stringr str_extract
.processMismatchLetters <- function(mismatches){
    mismatches <- strsplit(mismatches, split=",") 
    
    # Getting mismatch positions:
    mismatches.letters <- lapply(mismatches, function(x){
        x <- str_extract(x, "\\:[A,C,G,T]") 
        x <- gsub("\\:","",x)
        x <- as.character(x)
        for (i in seq_len(N_MAX_MISMATCHES-1)){
            if (length(x)==i){
                x <- c(x, rep(NA, N_MAX_MISMATCHES-i))
                break
            }
        }
        return(x)
    })
    mismatches.letters <- do.call("rbind",
                                  mismatches.letters)
    colnames(mismatches.letters) <- paste0("mmnuc",
                                           seq_len(N_MAX_MISMATCHES))
    return(mismatches.letters)
}

context('Testing runBowtie function')

# Creating bowtie index:
fasta  <- system.file(package="crisprBowtie","example/chr1.fa")
outdir <- tempdir()
Rbowtie::bowtie_build(fasta,
                      outdir=outdir,
                      force=TRUE,
                      prefix="tempIndex")
index <- file.path(outdir, "tempIndex")


test_that('Testing runBowtie', {
    seqs <- c("GGAAGTA", "GTGGACC", "GTGTGCG") 
    long_seq <- "AGGGAGGAGAGAGAGAGAGGGAGGTGGGTTAGGATTTAGGATTAGTTA"
    results_bowtie_0 <- runBowtie(seqs, bowtie_index=index, n_mismatches=0)
    results_bowtie_1 <- runBowtie(seqs, bowtie_index=index, n_mismatches=1)
    results_bowtie_2 <- runBowtie(seqs, bowtie_index=index, n_mismatches=2)
    results_bowtie_3 <- runBowtie(seqs, bowtie_index=index, n_mismatches=3)
    results_null <- runBowtie(long_seq, bowtie_index=index, n_mismatches=0)
    update <- FALSE

    expect_equal_to_reference(results_bowtie_0,
                              file=file.path("objects/results_bowtie_0.rds"),
                              update=update)
    expect_equal_to_reference(results_bowtie_1,
                              file=file.path("objects/results_bowtie_1.rds"),
                              update=update)
    expect_equal_to_reference(results_bowtie_2,
                              file=file.path("objects/results_bowtie_2.rds"),
                              update=update)
    expect_equal_to_reference(results_bowtie_3,
                              file=file.path("objects/results_bowtie_3.rds"),
                              update=update)
    expect_true(is.null(results_null))
})

---
title: "crisprBowtie: alignment of gRNA spacer sequences using bowtie"
author: 
- name: Jean-Philippe Fortin
  affiliation: Department of Data Science and Statistical Computing, gRED, 
   Genentech
  email: fortin946@gmail.com
date: "`r Sys.Date()`"
output: 
  BiocStyle::html_document:
    toc_float: true
#    theme: paper
    number_sections: true
vignette: >
  %\VignetteIndexEntry{Introduction to crisprBowtie}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
bibliography: references.bib
---



# Overview of crisprBowtie

`crisprBowtie` provides two main functions to align short DNA sequences to
a reference genome using the short read aligner bowtie [@langmead2009bowtie]
and return the alignments as R objects: `runBowtie` and `runCrisprBowtie`.
It utilizes the Bioconductor package `Rbowtie` to access the Bowtie program
in a platform-independent manner. This means that users do not need to install
Bowtie prior to using `crisprBowtie`. 


The latter function (`runCrisprBowtie`) is specifically designed
to map and annotate CRISPR guide RNA (gRNA) spacer sequences using
CRISPR nuclease objects and CRISPR genomic arithmetics defined in
the Bioconductor package
[crisprBase](https://github.com/crisprVerse/crisprBase).
This enables a fast and accurate on-target and off-target search of 
gRNA spacer sequences for virtually any type of CRISPR nucleases. 
It also provides an off-target search engine for our main gRNA design package [crisprDesign](https://github.com/crisprVerse/crisprDesign) of the 
[crisprVerse](https://github.com/crisprVerse) ecosystem. See the 
`addSpacerAlignments` function in `crisprDesign` for more details. 


# Installation and getting started

## Software requirements

### OS Requirements

This package is supported for macOS, Linux and Windows machines.
Package was developed and tested on R version 4.2.


## Installation from Bioconductor

`crisprBowtie` can be installed from from the Bioconductor devel branch
using the following commands in a fresh R session:

```{r, eval=FALSE}
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install(version="devel")
BiocManager::install("crisprBowtie")
```




# Building a bowtie index

To use `runBowtie` or `runCrisprBowtie`, users need to first build a Bowtie
genome index. For a given genome, this step has to be done only once. 
The `Rbowtie` package conveniently provides the function `bowtie_build`
to build a Bowtie index from any custom genome from a FASTA file.

As an example, we build a Bowtie index for a small portion of the human
chromosome 1 (`chr1.fa` file provided in the `crisprBowtie` package) and
save the index file as `myIndex` to a temporary directory:

```{r}
library(Rbowtie)
fasta <- file.path(find.package("crisprBowtie"), "example/chr1.fa")
tempDir <- tempdir()
Rbowtie::bowtie_build(fasta,
                      outdir=tempDir,
                      force=TRUE,
                      prefix="myIndex")
```

To learn how to create a Bowtie index for a complete genome or transcriptome,
please visit our [tutorial page](https://github.com/crisprVerse/Tutorials/tree/master/Building_Genome_Indices).



# Alignment using `runCrisprBowtie`

As an example, we align 6 spacer sequences (of length 20bp) to the
custom genome built above, allowing a maximum of 3 mismatches between the 
spacer and protospacer sequences. 

We specify that the search is for the wildtype Cas9 (SpCas9) nuclease
by providing the `CrisprNuclease` object `SpCas9` available through the 
`crisprBase` package. The argument `canonical=FALSE` specifies that 
non-canonical PAM sequences are also considered (NAG and NGA for SpCas9).
The function `getAvailableCrisprNucleases` in `crisprBase` returns a character
vector of available `crisprNuclease` objects found in `crisprBase`. 

```{r, warning=FALSE, message=FALSE}
library(crisprBowtie)
data(SpCas9, package="crisprBase")
crisprNuclease <- SpCas9
spacers <- c("TCCGCGGGCGACAATGGCAT",
             "TGATCCCGCGCTCCCCGATG",
             "CCGGGAGCCGGGGCTGGACG",
             "CCACCCTCAGGTGTGCGGCC",
             "CGGAGGGCTGCAGAAAGCCT",
             "GGTGATGGCGCGGGCCGGGC")
runCrisprBowtie(spacers,
                crisprNuclease=crisprNuclease,
                n_mismatches=3,
                canonical=FALSE,
                bowtie_index=file.path(tempDir, "myIndex"))
```



# Applications beyond CRISPR

The function `runBowtie` is similar to `runCrisprBowtie`,
but does not impose constraints on PAM sequences.
It can be used to search for any short read sequence in a genome.

## Example using RNAi (siRNA design)

Seed-related off-targets caused by mismatch tolerance outside of the
seed region is a well-studied and characterized problem observed in RNA
interference (RNA) experiments. `runBowtie` can be used to map shRNA/siRNA seed
sequences to reference genomes to predict putative off-targets:

```{r, eval=TRUE}
seeds <- c("GTAAAGGT", "AAGGATTG")
runBowtie(seeds,
          n_mismatches=2,
          bowtie_index=file.path(tempDir, "myIndex"))
```




# Reproducibility

```{r}
sessionInfo()
```



# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runBowtie.R
\name{runBowtie}
\alias{runBowtie}
\title{Perform short sequence alignment with bowtie}
\usage{
runBowtie(
  sequences,
  bowtie_index,
  bsgenome = NULL,
  n_mismatches = 0,
  all_alignments = TRUE,
  n_max_alignments = 1000,
  verbose = TRUE
)
}
\arguments{
\item{sequences}{Character vector of DNA sequences.}

\item{bowtie_index}{String specifying path to a bowtie index.}

\item{bsgenome}{BSgenome object.}

\item{n_mismatches}{Integer between 0 and 3 specifying maximum number
of mismatches allowed between query sequences and target DNA.
0 by default.}

\item{all_alignments}{Should all possible alignments be returned?
TRUE by default.}

\item{n_max_alignments}{Maximum number of alignments to return if
\code{all_alignments} is FALSE. 1000 by default.}

\item{verbose}{Should messages be printed to the console?
TRUE by default.}
}
\value{
A data.frame of the alignments with the following columns:
   \itemize{
       \item \code{query} — string specifying query DNA sequence
       \item \code{target} — string specifying target DNA sequence
       \item \code{chr} - string specifying chromosome name
       \item \code{pos} - string specifying genomic coordinate of the start
             of the target DNA sequence
       \item \code{strand} - string specifying strand ("+" or "-") 
       \item \code{n_mismatches} - integer specifying number of mismatches
             between query and target sequences
   }
}
\description{
Perform short sequence alignment with bowtie.
}
\details{
fasta <- system.file(package="crisprBowtie", "example/chr1.fa")
outdir <- tempdir()
Rbowtie::bowtie_build(fasta,outdir=outdir, force=TRUE, prefix="tempIndex")

\code{runBowtie} can be used to map short DNA sequences 
    to a reference genome. To search for sequences while imposing
    constraints on PAM sequences (such as gRNA spacer sequences), see
    \code{runCrisprBowtie} instead.
}
\examples{
fasta <- system.file(package="crisprBowtie", "example/chr1.fa")
outdir <- tempdir()
Rbowtie::bowtie_build(fasta,outdir=outdir, force=TRUE, prefix="tempIndex")
index <- file.path(outdir, "tempIndex")
seqs <- c("GGAAGT",
          "GTGGAC",
          "GTGTGC") 
results <- runBowtie(seqs, bowtie_index=index, n_mismatches=2)


}
\seealso{
\code{\link{runCrisprBowtie}} to map gRNA spacer sequences.
}
\author{
Jean-Philippe Fortin
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runCrisprBowtie.R
\name{runCrisprBowtie}
\alias{runCrisprBowtie}
\title{Perform CRISPR gRNA spacer alignment with bowtie}
\usage{
runCrisprBowtie(
  spacers,
  mode = c("protospacer", "spacer"),
  bowtie_index = NULL,
  bsgenome = NULL,
  crisprNuclease = NULL,
  canonical = TRUE,
  ignore_pam = FALSE,
  n_mismatches = 0,
  all_alignments = TRUE,
  n_max_alignments = 1000,
  force_spacer_length = FALSE,
  rna_strict_directionality = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{spacers}{Character vector specifying gRNA spacer sequences.
Sequences must all be of equal length.}

\item{mode}{String specifying which alignment mode should be used:
\code{protospacer} or \code{spacer}. 
For RNA-targeting nucleases such as CasRx, only the 
protospacer mode can be used.}

\item{bowtie_index}{String specifying path to a bowtie index.}

\item{bsgenome}{A \linkS4class{BSgenome} object.
Must be provided if \code{mode} is "spacer".
Ignore}

\item{crisprNuclease}{A \linkS4class{CrisprNuclease} object.}

\item{canonical}{Should only canonical PAM sequences be considered?
TRUE by default.}

\item{ignore_pam}{Should PAM sequences be ignore?
If TRUE, all alignments are returned regardless of PAM tolerance.
FALSE by default.}

\item{n_mismatches}{Integer between 0 and 3 specifying maximum number
of mismatches allowed between spacer sequences and target DNA.
0 by default.}

\item{all_alignments}{Should all possible alignments be returned?
TRUE by default.}

\item{n_max_alignments}{Maximum number of alignments to return if
\code{all_alignments} is FALSE. 1000 by default.}

\item{force_spacer_length}{Should the spacer length be overwritten in the
\code{crisprNuclease} object? FALSE by default.}

\item{rna_strict_directionality}{Should only protospacers found in the 
original direction of the RNA be considered for RNA-targeting
nucleases? TRUE by default.}

\item{verbose}{Should messages be printed to the console? TRUE by default.}
}
\value{
A data.frame of the spacer alignments with the following columns:
   \itemize{
       \item \code{spacer} — string specifying gRNA spacer sequence
       \item \code{protospacer} — string specifying target protospacer sequence
       \item \code{pam} — string specifying target PAM sequence
       \item \code{chr} - string specifying chromosome name
       \item \code{pam_site} - string specifying genomic coordinate of the
             first nucleotide of the PAM sequence.
       \item \code{strand} - string specifying strand ("+" or "-") 
       \item \code{n_mismatches} - integer specifying number of mismatches
             between spacer and protospacer sequences
       \item \code{canonical} - logical indicating whether or not PAM sequence
             is canonical. 
   }
}
\description{
Perform CRISPR gRNA spacer alignment with bowtie.
}
\details{
When \code{mode} is "spacer", spacer sequences are aligned to the
    reference index without appending PAM sequences first. This requires the
    specification of a \linkS4class{BSgenome} object through the argument
    \code{bsgenome} to validate that the aligned spacer sequences are
    adjacent to valid PAM sequences. 
    
    When \code{mode} is "protospacer", sequences are aligned with all
    valid PAM sequences appended (spacer + PAM). The set of valid PAM
    sequences depend on the inputs  \code{canonical} and \code{ignore_pam}.
    This is faster than the "spacer" mode if the number of possible
    PAM sequences is small (e.g. SpCas9).

    For RNA-targeting nucleases, such as RfxCas13d (CasRx), the bowtie
    index should be built on a transcriptome. For such applications,
    only the "protospacer" mode can be used as there is no
    corresponding bsgenome package. The protospacer sequences
    searched in the reference index will be the reverse complement
    of the input spacer sequences.
}
\examples{
fasta <- system.file(package="crisprBowtie", "example/chr1.fa")
outdir <- tempdir()
Rbowtie::bowtie_build(fasta,outdir=outdir, force=TRUE, prefix="tempIndex")
index <- file.path(outdir, "tempIndex")
seqs <- c("GGAAATTCCCCCAGTGGCGC",
          "ACACAGCTGCGGACAGGGCC")
data(SpCas9, package="crisprBase")
results <- runCrisprBowtie(seqs,
                           bowtie_index=index,
                           n_mismatches=2,
                           crisprNuclease=SpCas9)

}
\seealso{
\code{\link{runBowtie}} to map general DNA sequences.
}
\author{
Jean-Philippe Fortin
}
